"""Metrics collection classes."""

import pandas as pd
import requests
from loguru import logger as log

from prometheus_client import Summary


HTTP_REQUEST_TIME = Summary(
    "rlm_request_response_seconds", "time spent waiting for RLM to return data"
)
HTML_PARSING_TIME = Summary(
    "rlm_parse_data_seconds", "time spent parsing data from the RLM response"
)


@HTTP_REQUEST_TIME.time()
def send_http_request(url, data, timeout):
    """Wrapper for sending (and timing) an HTTP POST request.

    Parameters are identical to those of `requests.post()` with the same names.

    Exceptions are silenced into log messages as they shouldn't be passed on
    when running in service mode.

    Returns
    -------
    str or None
        The `text` property of the response object created by the `post()` call
        or `None` in case the call raised an exception.
    """
    log.trace(f"Sending HTTP POST: url=[{url}], data=[{data}]")
    try:
        response = requests.post(url=url, data=data, timeout=timeout)
    except Exception as err:  # pylint: disable-msg=broad-except
        log.error(f"Failed fetching data from RLM: {err}")
        return None

    log.trace(f"=-=-=-=-=\n{response.text}\n=-=-=-=-=")
    return response.text


@HTML_PARSING_TIME.time()
def parse_html_into_dataframes(html, header):
    """Wrapper for parsing the RLM data into a dataframe (and timing it).

    Exceptions are silenced into log messages as they shouldn't be passed on
    when running in service mode.

    Parameters
    ----------
    html : str
        The HTML containing the tables to be parsed by Pandas.
    header : int or list-like
        The row to use to make the column headers, passed on directly to the
        `read_html()` call.

    Returns
    -------
    list(pandas.DataFrame) or None
        A list of dataframe objects, one per table, or `None` in case parsing
        the data failed or the input was `None`.
    """
    if html is None:  # happens if the HTTP request failed
        return None
    try:
        tables = pd.read_html(html, header=header)
    except Exception as err:  # pylint: disable-msg=broad-except
        log.error(f"Failed parsing tables from HTML: {err}")
        return None

    return tables


class RlmCollector:
    """Abstract base collector class."""

    def __init__(self, config):
        # log.trace(f"Instantiating {self.__class__}...")
        self.base_uri = f"{config.rlm_uri}"
        log.debug(f"Using base URI: [{self.base_uri}]")
        self._uri = None
        self.postdata = None

    @property
    def uri(self):
        """Getter method for the `uri` attribute.

        Raises
        ------
        TypeError
            Raised in case the `uri` attribute is set to `None`.
        """
        if self._uri is None:
            raise TypeError("Instance doesn't have its `uri` attribute set!")
        return self._uri

    def set_stats_name(self, value):
        """Set name of the stats to request and process.

        This method actually sets the `uri` instance attribute depending on the
        `rlm_uri` configuration value.

        Parameters
        ----------
        value : str
            The name of the stats to request from RLM, e.g. `rlmstat_lic_process`.
        """
        if self.base_uri[0:5] == "http:":
            full_uri = f"{self.base_uri}/goform/{value}"
        else:
            full_uri = f"{self.base_uri}/{value}.html"
        log.trace(f"Setting `uri` value of {self.__class__} to [{full_uri}]...")
        self._uri = full_uri

    def collect(self):
        """Request metrics from RLM and parse them into a dataframe.

        Returns
        -------
        list[DataFrame]
        """
        log.trace(f"Collecting data from [{self.uri}]...")
        html = self.uri
        if html[0:4] == "http":  # it URI starts with 'http' request the data:
            html = send_http_request(url=self.uri, data=self.postdata, timeout=5)

        tables = parse_html_into_dataframes(html, header=0)

        return tables


class LicProcessCollector(RlmCollector):
    """Collector for "lic_process" data."""

    def __init__(self, config):
        log.trace(f"Instantiating {self.__class__}...")
        super().__init__(config)
        self.set_stats_name("rlmstat_lic_process")
        self.postdata = {
            "isv": config.isv,
            "instance": "0",
            "host": "",
            "wb": "rlmstat_lic",
            "pool": "0",
            "password": "",
            "ok": "Refresh",
        }



"""Command-line functions / entry points."""

import sys
from time import sleep

import click
from loguru import logger as log
from prometheus_client import Info, start_http_server

from . import __version__
from .config import get_config_from_env, load_config_file
from .metrics import RlmProductMetrics


def configure_logging(verbose: int):
    """Configure loguru logging / change log level.

    Parameters
    ----------
    verbose : int
        The desired log level, 0=WARNING (do not change the logger config),
        1=INFO, 2=DEBUG, 3=TRACE. Higher values will map to TRACE.
    """
    level = "WARNING"
    if verbose == 1:
        level = "INFO"
    elif verbose == 2:
        level = "DEBUG"
    elif verbose >= 3:
        level = "TRACE"
    # set up logging, loguru requires us to remove the default handler and
    # re-add a new one with the desired log-level:
    log.remove()
    log.add(sys.stderr, level=level)
    log.info(f"Set logging level to [{level}] ({verbose}).")


@click.command(help="Run the RLM metrics collector and exporter.")
@click.option("--config", type=str, help="A YAML configuration file.")
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Increase logging verbosity, may be repeated up to 3 times.",
)
def run_rlm_exporter(verbose, config):
    """Main CLI entry point for the RLM exporter. Blocking.

    Parameters
    ----------
    verbose : int
        Verbosity level for logging, ranges from 0 ("WARNING") to 3 ("TRACE").
    config : str
        A path to a configuration file. If `None` the settings will be derived
        from environment variables.
    """
    # do a first logging configuration to respect the command line parameters:
    configure_logging(verbose)

    if config:
        configuration = load_config_file(config)
    else:
        configuration = get_config_from_env()

    # verbosity might have been specified in the config / environment:
    if configuration.verbosity > verbose:
        configure_logging(configuration.verbosity)

    start_http_server(configuration.exporter_port)
    log.success(f"Providing metrics via HTTP on port {configuration.exporter_port}.")
    metrics = RlmProductMetrics(configuration)

    info = Info(
        name="rlm_exporter",
        documentation="Information on the RLM metrics collector and exporter service",
    )
    info.info(
        {
            "version": __version__,
            "collection_interval": f"{configuration.interval}s",
        }
    )

    log.success(
        f"{__package__} {__version__} started, "
        f"collection interval {configuration.interval}s."
    )

    while True:
        log.trace("Updating pool status...")
        try:
            metrics.update_metrics()
        except Exception as err:  # pylint: disable-msg=broad-except
            log.error(f"Updating metrics failed: {err}")
        sleep(configuration.interval)



"""Configuration loader function(s)."""

import os
from box import Box


def load_config_file(filename):
    """Assemble a config object by loading values from a file."""
    config = Box.from_yaml(filename=filename)
    if "rlm_uri" not in config.keys():
        config.rlm_uri = "http://localhost:5054"
    if "exporter_port" not in config.keys():
        config.exporter_port = 8909
    if "interval" not in config.keys():
        config.interval = 60
    if "checkout_details" not in config.keys():
        config.checkout_details = False
    if "verbosity" not in config.keys():
        config.verbosity = 0
    if "isv" not in config.keys():
        raise ValueError("Config is missing setting for 'isv'!")
    return config


def get_config_from_env():
    """Assemble a config object from envrionment variables."""
    try:
        isv = os.environ["RLM_ISV"]
    except Exception as err:
        raise KeyError("Error getting configuration values:", err) from err

    rlm_uri = os.environ.get("RLM_URI", "http://localhost:5054")
    exporter_port = int(os.environ.get("RLM_EXPORTER_PORT", 8909))
    interval = int(os.environ.get("RLM_EXPORTER_INTERVAL", 60))
    ignoreproducts = os.environ.get("RLM_IGNORE_PRODUCTS", None)
    checkout_details = bool(os.environ.get("RLM_CHECKOUT_DETAILS", False))
    verbosity = int(os.environ.get("RLM_EXPORTER_VERBOSITY", 0))

    return Box(
        {
            "rlm_uri": rlm_uri,
            "exporter_port": exporter_port,
            "interval": interval,
            "isv": isv,
            "ignoreproducts": ignoreproducts,
            "checkout_details": checkout_details,
            "verbosity": verbosity,
        }
    )
